#!/usr/bin/env python
# -*- encoding: utf-8 -*-
##
## Copyright © 2012 Thibaut Girka <thib@sitedethib.com>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation; version 3 only.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##

import sys
import pygame
import math

from ptis.resource.loader import Loader
from ptis.formats.level import Level


WIDTH, HEIGHT = 216, 144
FPS = 25


class Editor(object):
    def __init__(self, level_path=None, zoomlevel=4, fullscreen=False):
        self.zoomlevel = zoomlevel
        self.level = None
        self.level_background = None
        self.level_path = level_path
        self.hide_background = False
        self.running = False

        loader = Loader()


        width = WIDTH * zoomlevel + 300
        height = HEIGHT * zoomlevel + 20
        flags = 0
        if fullscreen:
            flags |= pygame.FULLSCREEN

        self.window = pygame.display.set_mode((width, height), flags)
        pygame.display.set_caption(('%s - editor' % level_path) if level_path else 'editor')

        if self.level_path:
            with open(self.level_path, 'rb') as file:
                self.level = Level.read(file)
            background = loader.load_level_backgrounds(self.level)[0]
            width, height = background.get_size()
            self.level_background = pygame.transform.scale(background,
                                                           (width * zoomlevel,
                                                            height * zoomlevel))
        else:
            self.level = Level('') #TODO

        self.selected_polygon = None
        self.selected_point = None
        self.map_x = 0


    def run(self):
        self.running = True
        clock = pygame.time.Clock()
        while self.running:
            clock.tick(FPS)
            self.handle_events()
            self.draw()


    def update_map_x(self, delta):
        self.map_x += delta
        if self.level_background:
            max_map_x = self.level_background.get_width() // self.zoomlevel
            max_map_x -= WIDTH
        else:
            max_map_x = 0
        self.map_x = max(min(max_map_x, self.map_x), 0)


    def handle_map_click(self, event):
        x, y = event.pos
        x //= self.zoomlevel
        x += self.map_x
        y //= self.zoomlevel
        if event.button == 1:
            if not self.selected_polygon:
                self.selected_polygon = (0, [])
                self.level.polygons.append(self.selected_polygon)
            points = self.selected_polygon[1]
            index = points.index(self.selected_point) if self.selected_point in points else 0
            self.selected_point = [x, y]
            points.insert(index + 1, self.selected_point)
            pygame.display.set_caption(('*%s - editor' % self.level_path)
                                            if self.level_path else 'editor')
        elif event.button == 3:
            for polygon in self.level.polygons:
                if self.point_in_polygon((x, y), polygon[1]):
                    self.selected_polygon = polygon
            if self.selected_polygon:
                for point in self.selected_polygon[1]:
                    if not self.selected_point:
                        self.selected_point = point
                    else:
                        x1, y1 = point
                        x2, y2 = self.selected_point
                        if math.hypot(x - x1, y - y1) < math.hypot(x - x2, y - y2):
                            self.selected_point = point


    def handle_events(self):
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self.running = False
            elif event.type == pygame.KEYDOWN:
                if event.key == pygame.K_s and self.level_path:
                    with open(self.level_path, 'wb') as file:
                        self.level.write(file)
                    pygame.display.set_caption('%s - editor' % self.level_path)
                elif event.key == pygame.K_u:
                    self.selected_polygon = None
                    self.selected_point = None
                elif event.key in (pygame.K_ESCAPE, pygame.K_q):
                    self.running = False
                elif event.key in (pygame.K_UP, pygame.K_DOWN, pygame.K_LEFT, pygame.K_RIGHT):
                    if self.selected_point:
                        comp, direction = {pygame.K_UP: (1, -1),
                                           pygame.K_DOWN: (1, 1),
                                           pygame.K_LEFT: (0, -1),
                                           pygame.K_RIGHT: (0, 1)}[event.key]
                        self.selected_point[comp] += direction
                        pygame.display.set_caption(('*%s - editor' % self.level_path)
                                if self.level_path else 'editor')
                elif event.key == pygame.K_TAB:
                    if self.selected_polygon:
                        index = self.level.polygons.index(self.selected_polygon)
                        self.selected_polygon = self.level.polygons[(index + 1) % len(self.level.polygons)]
                    elif self.level.polygons:
                        self.selected_polygon = self.level.polygons[0]

                    if self.selected_polygon and self.selected_polygon[1]:
                        self.selected_point = self.selected_polygon[1][0]
                    else:
                        self.selected_point = None
                elif event.key == pygame.K_DELETE and self.selected_polygon:
                    if self.selected_polygon:
                        points = self.selected_polygon[1]
                        if self.selected_point in points:
                            _index = points.index(self.selected_point)
                            prev_point = points[(_index - 1) % len(points)]
                        else:
                            prev_point = None
                    if self.selected_point:
                        self.selected_polygon[1].remove(self.selected_point)
                        pygame.display.set_caption(('*%s - editor' % self.level_path)
                                                    if self.level_path else 'editor')
                    if self.selected_polygon[1]:
                        self.selected_point = prev_point
                    else:
                        self.selected_point = None
                        self.level.polygons.remove(self.selected_polygon)
                        self.selected_polygon = None
            elif event.type == pygame.MOUSEBUTTONDOWN:
                x, y = event.pos
                if event.button == 4:
                    self.update_map_x(-10)
                elif event.button == 5:
                    self.update_map_x(+10)
                elif x < WIDTH * self.zoomlevel and y < HEIGHT * self.zoomlevel:
                    self.handle_map_click(event)


    def point_in_polygon(self, point, points):
        x, y = point
        nb_points = len(points)
        oldpoint = points[-1]

        inside = False

        for vert in points:
            x1, y1 = oldpoint if vert[0] > oldpoint[0] else vert
            x2, y2 = vert if vert[0] > oldpoint[0] else oldpoint

            if (((x1 < x) == (x <= x2))
                and (y - y1) * (x2 - x1) < (y2 - y1) * (x - x1)):
                inside = not inside
            oldpoint = vert

        return inside


    def draw(self):
        if self.selected_polygon:
            points = self.selected_polygon[1]
            if self.selected_point in points:
                _index = points.index(self.selected_point)
                next_point = points[(_index + 1) % len(points)]
            else:
                next_point = None
        else:
            points = None
            next_point = None

        self.window.fill((0, 0, 255))
        left = self.map_x * self.zoomlevel
        if self.level_background and not self.hide_background:
            self.window.blit(self.level_background, (0, 0),
                             (left, 0,
                              WIDTH * self.zoomlevel, HEIGHT * self.zoomlevel))
        if self.level:
            for polygon in self.level.polygons:
                polygon_type, points = polygon
                color = (255, 255, 255) #TODO
                points2 = [(x * self.zoomlevel - left, y * self.zoomlevel) for x, y in points]
                pygame.draw.polygon(self.window, color, points2 + [points2[0]], 1)
                for orig, point in zip(points, points2):
                    if orig is self.selected_point:
                        color2 = (0, 255, 0)
                    elif orig is next_point:
                        color2 = (255, 0, 0)
                    else:
                        color2 = (255, 255, 255)
                    pygame.draw.circle(self.window, color2, point, 4)

        # Info box
        self.window.fill((255, 255, 255),
                         (WIDTH * self.zoomlevel, 0, 300, HEIGHT * self.zoomlevel + 20))
        #TODO

        # Scroll
        if self.level_background:
            self.window.fill((130, 130, 130),
                             (0, HEIGHT * self.zoomlevel, WIDTH * self.zoomlevel, 20))
            right = WIDTH * self.zoomlevel
            pygame.draw.polygon(self.window, (255, 255, 255), [(8, HEIGHT * self.zoomlevel + 10),
                                                               (15, HEIGHT * self.zoomlevel + 5),
                                                               (15, HEIGHT * self.zoomlevel + 15)], 0)
            pygame.draw.polygon(self.window, (255, 255, 255), [(right - 8, HEIGHT * self.zoomlevel + 10),
                                                               (right - 15, HEIGHT * self.zoomlevel + 5),
                                                               (right - 15, HEIGHT * self.zoomlevel + 15)], 0)
            min_x = 20
            max_x = right - 20
            max_length = max_x - min_x
            ratio = float(max_length) / (self.level_background.get_width() // self.zoomlevel)

            pygame.draw.rect(self.window, (255, 255, 255), (min_x + ratio * self.map_x,
                                                            HEIGHT * self.zoomlevel + 6,
                                                            ratio * WIDTH,
                                                            9))


        pygame.display.flip()


pygame.display.init()
pygame.font.init()
editor = Editor(sys.argv[1] if len(sys.argv) > 1 else None)
editor.run()

