# -*- encoding: utf-8 -*-
##
## Copyright © 2012 Thibaut Girka <thib@sitedethib.com>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation; version 3 only.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##

from math import hypot, atan2, sqrt

INF = float('inf')
EPSILON = 0.05
SQRT2_2 = sqrt(2) / 2.

def reorder_clockwise(points):
    '''
    Return the list of points for a given polygon such that the outline is
    counter-clockwise.
    '''
    cx = sum(x for x, y in points) / len(points)
    cy = sum(y for x, y in points) / len(points)
    angle = 0
    oldx, oldy = points[-1]
    for x, y in points:
        ax, ay = oldx - cx, oldy - cy
        bx, by = x - cx, y - cy
        angle += atan2(ax * by - ay * bx, ax * bx + ay * by)
        oldx, oldy = x, y
    return list(reversed(points)) if angle < 0 else points



class Polygon(object):
    class Extrema(dict):
        def __init__(self, polygon):
            self.polygon = polygon


        def __missing__(self, normal):
            nx, ny = normal
            points = [nx * x + ny * y for x, y in self.polygon.points]
            self[normal] = (min(points), max(points))
            return self[normal]


    def __init__(self, points):
        points = reorder_clockwise(points)
        self.points = points
        self.center = sum(x for x, y in points) / len(points), sum(y for x, y in points) / len(points)
        self.bounding_box = [min(x for x, y in points),
                             min(y for x, y in points),
                             max(x for x, y in points),
                             max(y for x, y in points)]
        self.segments = [(points[i], points[(i+1) % len(points)])
                            for i in xrange(len(points))]
        self.vecs = [(b[0] - a[0], b[1] - a[1]) for a, b in self.segments]
        self.normals = [(y / hypot(x, y), - x / hypot(x, y))
                            for x, y in self.vecs]
        self.extrema = Polygon.Extrema(self)



class Collidable(object):
    def __init__(self, polygon, pos, friction=0.6, bounce=0.0):
        self.polygon = polygon
        self.x, self.y = pos
        self.speed = [0, 0]
        self.bounding_box = [0, 0, 0, 0]
        self.friction = friction
        self.bounce = bounce
        self.on_ground = False
        self.removed = False


    def update_bounding_box(self, time=1.):
        minx, miny, maxx, maxy = self.polygon.bounding_box
        minx = self.x + min(minx + self.speed[0] * time, minx)
        miny = self.y + min(miny + self.speed[1] * time, miny)
        maxx = self.x + max(maxx + self.speed[0] * time, minx)
        maxy = self.y + max(maxy + self.speed[1] * time, maxy)
        self.bounding_box = minx, miny, maxx, maxy


    def update(self, time):
        self.x += self.speed[0] * time
        self.y += self.speed[1] * time


    def on_collide(self, collision):
        ellapsed_time, (a, b), (nx, ny) = collision
        if ny <= -SQRT2_2:
            self.on_ground = True
        speed = self.speed[0] - b.speed[0], self.speed[1] - b.speed[1]
        l = nx * self.speed[0] + ny * self.speed[1]
        vn = nx * l, ny * l
        vt = self.speed[0] - vn[0], self.speed[1] - vn[1]
        bounce = (a.bounce + b.bounce) / 2.
        friction = (a.friction + b.friction) / 2.
        self.speed = [vt[0] * friction - vn[0] * bounce,
                      vt[1] * friction - vn[1] * bounce]


    def check_upcoming_collisions(self, entities):
        if entities:
            return min(self.preventive_collision_check(entity)
                        for entity in entities if entity is not self)
        else:
            return (INF, (None, None), (0, 0))


    def collision_check(self, entity, time):
        dx_self, dy_self = self.x + self.speed[0] * time, self.y + self.speed[1] * time
        dx_other, dy_other = entity.x + entity.speed[0] * time, entity.y + entity.speed[1] * time
        polygon, other_polygon = self.polygon, entity.polygon

        dx = (other_polygon.center[0] + dx_other) - (polygon.center[0] + dx_self)
        dy = (other_polygon.center[1] + dy_other) - (polygon.center[1] + dy_self)

        for polygon, (dx, dy) in ((polygon, (dx, dy)),
                                  (other_polygon, (-dx, -dy))):
            for nx, ny in polygon.normals:
                if dx * nx + dy * ny < 0:
                    continue

                points_self = self.polygon.extrema[(nx, ny)]
                points_other = entity.polygon.extrema[(nx, ny)]
                offset_self = nx * dx_self + ny * dy_self
                offset_other = nx * dx_other + ny * dy_other

                if (points_self[1] + offset_self + EPSILON < points_other[0] + offset_other
                    or points_other[1] + offset_other + EPSILON < points_self[0] + offset_self):
                    return False
        return True


    def preventive_collision_check(self, entity):
        # http://content.gpwiki.org/index.php/Polygon_Collision
        next_collision = (INF, (None, None), (0, 0))
        colliding = True

        minx, miny, maxx, maxy = self.bounding_box
        minx2, miny2, maxx2, maxy2 = entity.bounding_box

        if (maxx2 < minx) or (maxx < minx2) or (maxy2 < miny) or (maxy < miny2):
            return next_collision

        polygon, other_polygon = self.polygon, entity.polygon
        dx = (other_polygon.center[0] + entity.x) - (polygon.center[0] + self.x)
        dy = (other_polygon.center[1] + entity.y) - (polygon.center[1] + self.y)

        for polygon, (dx, dy) in ((polygon, (dx, dy)),
                                  (other_polygon, (-dx, -dy))):
            for nx, ny in polygon.normals:
                if dx * nx + dy * ny < 0:
                    continue

                if polygon is self.polygon:
                    normal = -nx, -ny
                else:
                    normal = nx, ny

                points_self = self.polygon.extrema[(nx, ny)]
                points_other = entity.polygon.extrema[(nx, ny)]
                speed_self = nx * self.speed[0] + ny * self.speed[1]
                speed_other = nx * entity.speed[0] + ny * entity.speed[1]
                offset_self = nx * self.x + ny * self.y
                offset_other = nx * entity.x + ny * entity.y

                if points_self[1] + offset_self < points_other[0] + offset_other:
                    colliding = False
                    speed = speed_self - speed_other
                    if speed == 0:
                        continue
                    distance = (points_other[0] + offset_other) - (points_self[1] + offset_self)
                    time = distance / speed
                    if time > 0 and self.collision_check(entity, time):
                        time = max(0., (distance - EPSILON) / speed)
                        next_collision = (time, (self, entity), normal)
                elif points_other[1] + offset_other < points_self[0] + offset_self:
                    colliding = False
                    speed = speed_other - speed_self
                    if speed == 0:
                        continue
                    distance = (points_self[0] + offset_self) - (points_other[1] + offset_other)
                    time = distance / speed
                    if time > 0 and self.collision_check(entity, time):
                        time = max(0., (distance - EPSILON) / speed)
                        next_collision = (time, (self, entity), normal)

        if colliding:
            raise Exception #TODO

        return next_collision
