#!/usr/bin/env python

import os
import pygame
from math import atan2, cos, sin, hypot
from ptis.formats.level import Level
from ptis.game.collidable import Collidable, Polygon
from ptis.game.game import Game
from ptis.resource.loader import Loader
import random

WIDTH, HEIGHT = 216, 144

class Player(Collidable):
    def __init__(self, pos, image):
        size = image.get_size()
        polygon = Polygon([(0, 7), (7, 0), (size[0] - 7, 0), (size[0], 7),
                           (size[0], size[1] - 10), (size[0] - 2, size[1]),
                           (2, size[1]), (0, size[1] - 10)])
        Collidable.__init__(self, polygon, pos, friction=0.9, bounce=0.)
        self.life = 10
        self.image = image.convert_alpha()
        self.size = size
        self.walking = 0
        self.direction = 0
        self.controllable = True
        self.holding_potion = True


    def handle_event(self, event):
        if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_SPACE and self.on_ground and self.controllable:
                self.speed[1] = -1.9
            elif event.key == pygame.K_LEFT:
                self.walking |= 1
            elif event.key == pygame.K_RIGHT:
                self.walking |= 2
        elif event.type == pygame.KEYUP:
            if event.key == pygame.K_LEFT:
                self.walking &= ~1
            elif event.key == pygame.K_RIGHT:
                self.walking &= ~2


    def tick(self, frame):
        if not self.controllable:
            return

        delta = 0.8 if self.on_ground else 0.05
        if self.walking & 1 and self.speed[0] > -2.:
            self.speed[0] = max(-2., self.speed[0] - delta)
            self.direction = 0
        elif self.walking & 2 and self.speed[0] < 2.:
            self.speed[0] = min(2., self.speed[0] + delta)
            self.direction = 1


    def draw(self, screen, zoomlevel, screen_rect):
        rect = (self.x * zoomlevel - screen_rect[0],  self.y * zoomlevel,
                self.size[0] * zoomlevel, self.size[1] * zoomlevel)
        image = pygame.transform.scale(self.image, rect[2:])
        if self.direction == 1:
            image = pygame.transform.flip(image, True, False)
        screen.blit(image, rect)



class Hamster(Collidable):
    def __init__(self, pos, image):
        w, h = image.get_size()
        size = w // 3, h
        polygon = Polygon([(0, 1), (2, 0), (size[0] - 1, 0), (size[0], 1),
                           (size[0], size[1] - 1), (size[0] - 2, size[1]),
                           (2, size[1]), (0, size[1] - 2)])
        Collidable.__init__(self, polygon, pos, friction=1., bounce=0.05)
        self.image = image.convert_alpha()
        self.size = size
        self.walking = 0
        self.direction = 0
        self.step = 0


    def tick(self, frame):
        if frame % 3 == 0:
            self.step = (self.step + 1) % 3
        delta = 0.8 if self.on_ground else 0.05
        if self.walking & 1 and self.speed[0] > -2.5:
            self.speed[0] = max(-2.5, self.speed[0] - delta)
            self.direction = 0
        elif self.walking & 2 and self.speed[0] < 2.5:
            self.speed[0] = min(2.5, self.speed[0] + delta)
            self.direction = 1


    def draw(self, screen, zoomlevel, screen_rect):
        rect = (self.x * zoomlevel - screen_rect[0],  self.y * zoomlevel,
                self.size[0] * zoomlevel, self.size[1] * zoomlevel)
        step = self.step if self.walking else 0
        image = pygame.transform.scale(self.image, (rect[2] * 3, rect[3]))
        if self.direction == 0:
            image = pygame.transform.flip(image, True, False)
        screen.blit(image, rect, (rect[2] * step, 0) + rect[2:])


class HealthBar(object):
    def __init__(self, image):
        self.image = image
        self.rects = []


    def update(self, life):
        self.rects = [None] * 5
        for i in xrange(5):
            if life <= i * 2:
                self.rects[i] = (62, 2, 4, 13)
            elif life <= i * 2 + 1:
                self.rects[i] = (12 * i + 8, 2, 4, 13)
            else:
                self.rects[i] = (12 * i + 2, 2, 4, 13)


    def draw(self, screen, zoomlevel):
        size = self.image.get_size()
        image = pygame.transform.scale(self.image, (size[0] * zoomlevel, size[1] * zoomlevel))
        x, y = 4, 4
        for rect in self.rects:
            rect = tuple(r * zoomlevel for r in rect)
            screen.blit(image, (x, y), rect)
            x += 4 + rect[2]



class TextBox(object):
    def __init__(self, image):
        self.size = image.get_size()
        self.x, self.y = (WIDTH - self.size[0]) // 2, HEIGHT - self.size[1] - 5
        self.image = image
        self.font = pygame.font.Font('data/font.pcf', 16)
        self.text = ''


    def draw(self, screen, zoomlevel):
        if not self.text:
            return
        rect = (self.x * zoomlevel, self.y * zoomlevel,
                self.size[0] * zoomlevel, self.size[1] * zoomlevel)
        image = self.image.copy()
        for i, line in enumerate(self.text.split('\n')):
            image.blit(self.font.render(line, False, (255, 255, 255)), (5, 4 + 10 * i))
        image = pygame.transform.scale(image, rect[2:])
        screen.blit(image, rect)



class Boss(Collidable):
    def __init__(self, pos, image):
        size = image.get_size()
        polygon = Polygon([(size[0] / 2., size[1] / 4.),
                           (3 * size[0] / 4., size[1] / 2.),
                           (size[0] / 2., 3 * size[1] / 4.),
                           (size[0] / 4., 3 * size[1] / 4.)])
        Collidable.__init__(self, polygon, pos)
        self.image = image
        self.size = image.get_size()
        self.speed[0] = 2.5
        self.speed[1] = 0.
        self.mass = 0.0
        self.life = 80


    def tick(self, frame):
        self.speed[1] = 0.
        if frame % 200 == 0:
            self.speed[0] *= -1.


    def draw(self, screen, zoomlevel, screen_rect):
        rect = (self.x * zoomlevel - screen_rect[0],  self.y * zoomlevel,
                self.size[0] * zoomlevel, self.size[1] * zoomlevel)
        image = pygame.transform.scale(self.image, rect[2:])
        screen.blit(image, rect)



class Particle(Collidable):
    def __init__(self, pos, size=4, lifetime=200, color=(0x5e, 0xa9, 0x22)):
        polygon = Polygon([(0, size / 2.), (size / 2., 0), (size, size / 2.), (size / 2, size)])
        Collidable.__init__(self, polygon, pos, friction=1.)
        self.color = color
        self.size = size
        self.lifetime = lifetime


    def tick(self, frame):
        pass


    def update(self, time):
        Collidable.update(self, time)
        if self.lifetime <= 0:
            self.removed = True
        self.lifetime -= time


    def draw(self, screen, zoomlevel, screen_rect):
        rect = (self.x * zoomlevel - screen_rect[0],  self.y * zoomlevel,
                self.size * zoomlevel, self.size * zoomlevel)
        pygame.draw.rect(screen, self.color, rect)



class UserBullet(Collidable):
    def __init__(self, pos, size=2, lifetime=80, color=(0x5e, 0xa9, 0x22)):
        polygon = Polygon([(0, size / 2.), (size / 2., 0), (size, size / 2.), (size / 2, size)])
        Collidable.__init__(self, polygon, pos, friction=1.)
        self.color = color
        self.size = size
        self.lifetime = lifetime


    def tick(self, frame):
        pass


    def update(self, time):
        Collidable.update(self, time)
        if self.lifetime <= 0:
            self.removed = True
        self.lifetime -= time


    def draw(self, screen, zoomlevel, screen_rect):
        rect = (self.x * zoomlevel - screen_rect[0],  self.y * zoomlevel,
                self.size * zoomlevel, self.size * zoomlevel)
        pygame.draw.rect(screen, self.color, rect)



class Bullet(Collidable):
    def __init__(self, boss):
        size = 4
        lifetime = 200
        w, h = boss.image.get_size()
        pos = boss.x + w // 2, boss.y + h // 2
        polygon = Polygon([(0, size / 2.), (size / 2., 0), (size, size / 2.), (size / 2, size)])
        Collidable.__init__(self, polygon, pos, friction=1., bounce=1.)
        self.speed[0] = boss.speed[0] * 0.5
        self.size = size
        self.lifetime = lifetime
        self.mass = 0.4


    def tick(self, frame):
        pass


    def update(self, time):
        Collidable.update(self, time)
        if self.lifetime <= 0:
            self.removed = True
        self.lifetime -= time


    def draw(self, screen, zoomlevel, screen_rect):
        rect = (self.x * zoomlevel - screen_rect[0],  self.y * zoomlevel,
                self.size * zoomlevel, self.size * zoomlevel)
        pygame.draw.rect(screen, (255, 10, 10), rect)



class GameRunner(object):
    def __init__(self, loader, zoomlevel):
        self.loader = loader
        self.zoomlevel = zoomlevel
        self.background = None
        self.parallax_background = None
        self.game = None
        self.healthbar = None
        self.textbox = TextBox(loader.load_image('textbox.png'))
        self.hamster = None
        self.boss = None

        loader.preload(('test', 'test2', 'out1', 'out2'))
        self.change_level('test')


    def get_screen_rect(self, zoomlevel):
        width, height = WIDTH, HEIGHT
        x = self.game.player.x - WIDTH / 2.
        x = min(x, self.background.get_width() // zoomlevel - WIDTH)
        x = max(0, x)
        y = 0
        return (x * zoomlevel, y * zoomlevel, width * zoomlevel, height * zoomlevel)


    def change_level(self, level_name, clear=True):
        self.frame = 0
        level = self.loader.load_level(level_name)
        if clear:
            player = Player((80, 72), loader.load_image('mad-scientist.png'))
            self.game = Game(level, player)
        else:
            self.game.regen(level)

        background, parallax_background = self.loader.load_level_backgrounds(level)
        width, height = background.get_size()
        self.background = pygame.transform.scale(background, (width * self.zoomlevel, height * self.zoomlevel))
        if parallax_background:
            width, height = parallax_background.get_size()
            self.parallax_background = pygame.transform.scale(parallax_background,
                                                              (width * self.zoomlevel,
                                                               height * self.zoomlevel))
        else:
            self.parallax_background = None


    def event_hooks(self, event, frame):
        if not hasattr(self, 'last_fired_frame'):
            return
        if event.type == pygame.KEYDOWN and event.key == pygame.K_LSHIFT:
            self.last_fired_frame = frame


    def draw_hooks(self, frame):
        if self.game.level is self.loader.level_cache['test']:
            if frame < 1000:
                surface.fill((0, 0, 0, 255))
                screen.blit(surface, (0, 0))
            elif frame < 1050:
                c = 255 - 255 * (frame - 1000) // (1050 - 1000)
                surface.fill((0, 0, 0, c))
                screen.blit(surface, (0, 0))
        elif self.game.level is self.loader.level_cache['test2']:
            if self.game.player.x > 290:
                surface.fill((0, 0, 0, 255))
                screen.blit(surface, (0, 0))
            elif self.game.player.x > 200:
                c = 255 * (self.game.player.x - 200) // (290 - 200)
                surface.fill((0, 0, 0, c))
                screen.blit(surface, (0, 0))
        elif self.game.level is self.loader.level_cache['out1']:
            if frame < 20:
                surface.fill((0, 0, 0, 255))
                screen.blit(surface, (0, 0))
            elif frame < 150:
                c = 255 - 255 * (frame - 20) // (150 - 20)
                surface.fill((0, 0, 0, c))
                screen.blit(surface, (0, 0))


    def update_hooks(self, frame):
        if frame % 12 == 0 and self.game.player.holding_potion:
            xdelta = 8 if self.game.player.direction else (24 - 8)
            particle = Particle((self.game.player.x + xdelta, self.game.player.y + 18), 1, 50)
            particle.bounce = 1.
            particle.speed[1] = -1.
            particle.speed[0] = random.uniform(-1, 1.)
            self.game.entities.append(particle)
        if self.game.level is self.loader.level_cache['test']:
            if frame == 0:
                self.game.player.controllable = False
            elif frame == 50:
                self.textbox.text = "Once upon a time lived\na mad scientist..."
            elif frame == 200:
                self.textbox.text = "No one ever understood\nthe poor man."
            elif frame == 350:
                self.textbox.text = "As a result, he was\nalways working alone."
            elif frame == 500:
                self.textbox.text = "And nobody ever\npaid him."
            elif frame == 650:
                self.textbox.text = "One day, he found a\nstrange creature."
            elif frame == 800:
                self.textbox.text = "And he decided to keep\nit and study it."
            elif frame == 1000:
                self.textbox.text = ""
                self.game.player.controllable = True
            elif frame == 1100:
                self.change_level('test2', False)
                self.game.player.image = loader.load_image('mad-scientist-dark.png')
        elif self.game.level is self.loader.level_cache['test2']:
            if frame == 8:
                self.game.player.controllable = False
                self.game.player.walking = 0
                self.textbox.text = "Scientist:\n?!"
            elif frame == 180:
                self.textbox.text = ""
            elif frame == 200:
                self.textbox.text = "Scientist:\nOh, right... I forgot"
            elif frame == 250:
                self.textbox.text = "Scientist:\nto pay the bills..."
            elif frame == 350:
                self.textbox.text = ""
            elif frame == 360:
                self.textbox.text = "Scientist:\nI must find a way to"
            elif frame == 410:
                self.textbox.text = "Scientist:\nbring the power back."
            elif frame == 510:
                self.textbox.text = ""
                self.game.player.controllable = True

            if self.game.player.x > 300:
                self.change_level('out1')
        elif self.game.level is self.loader.level_cache['out1']:
            if not self.hamster:
                self.hamster = Hamster((120, 80), loader.load_image('hamster.png'))
                self.hamster.direction = 1
                self.game.entities.append(self.hamster)

            if self.hamster.x - self.game.player.x < 50:
                self.hamster.walking = 2
            elif self.hamster.x - self.game.player.x > 100:
                self.hamster.walking = 0

            if frame <= 1:
                self.game.player.controllable = False
            elif frame == 100:
                self.textbox.text = "Scientist:\nA hamster!"
            elif frame == 150:
                self.textbox.text = "Scientist:\nThis is perfect! Good"
            elif frame == 200:
                self.textbox.text = "Scientist:\nthing I have a wheel!"
            elif frame == 250:
                self.textbox.text = "Scientist:\nIt should be able"
            elif frame == 300:
                self.textbox.text = "Scientist:\nto provide enough"
            elif frame == 350:
                self.textbox.text = "Scientist:\npower to lead my"
            elif frame == 400:
                self.textbox.text = "Scientist:\nproject to completion!"
            elif frame == 500:
                self.textbox.text = ""
                self.game.player.controllable = True
            elif frame > 500 and self.game.player.x > 40:
                self.textbox.text = ""
            elif frame > 500:
                self.textbox.text = "Scientist:\nChase the hamster!"

            if self.game.player.x > 520:
                self.change_level('out2', False)
                self.hamster.walking = 1
        elif self.game.level is self.loader.level_cache['out2']:
            if frame % 200 == 0:
                self.hamster.walking = 3 - self.hamster.walking
            if frame == 10:
                self.boss = Boss((105, 1), self.loader.load_image('boss.png'))
                self.game.entities.append(self.boss)
                self.healthbar = HealthBar(self.loader.load_image('life.png'))
                self.textbox.text = "Scientist:\nSomething's strange..."
            elif frame == 60:
                self.textbox.text = "Scientist:\nPress [SHIFT]!"
                self.last_fired_frame = 0
            elif frame == 120:
                self.textbox.text = ""

            if frame > 10 and frame % 50 == 0:
                bullet = Bullet(self.boss)
                self.game.entities.append(bullet)

            for entity in self.game.entities:
                if isinstance(entity, Bullet) and entity.collision_check(self.game.player, 0):
                    entity.removed = True
                    self.game.player.life -= 1
                    pos = entity.x, entity.y
                    dx, dy = entity.x - self.game.player.x, entity.y - self.game.player.y
                    dx += entity.polygon.center[0] - self.game.player.polygon.center[0]
                    dy += entity.polygon.center[1] - self.game.player.polygon.center[1]
                    angle = atan2(dy, dx)
                    for i in xrange(3):
                        for j in xrange(6):
                            particle = Particle(pos, 1, 80, color=(255, 0, 0))
                            particle.speed[0] = (1. + i / 3.) * cos(angle - 0.05 + 0.1 * j / 6.)
                            particle.speed[1] = (1. + i / 3.) * sin(angle - 0.05 + 0.1 * j / 6.)
                            self.game.entities.append(particle)

            for entity in self.game.entities:
                if isinstance(entity, UserBullet) and entity.collision_check(self.boss, 0):
                    entity.removed = True
                    self.boss.life -= 1
                    pos = entity.x, entity.y
                    dx, dy = entity.x - self.boss.x, entity.y - self.boss.y
                    dx += entity.polygon.center[0] - self.boss.polygon.center[0]
                    dy += entity.polygon.center[1] - self.boss.polygon.center[1]
                    angle = atan2(dy, dx)
                    for i in xrange(3):
                        for j in xrange(4):
                            particle = Particle(pos, 1, 30, color=(0, 255, 100))
                            particle.speed[0] = (1. + i / 3.) * cos(angle - 0.05 + 0.1 * j / 4.)
                            particle.speed[1] = (1. + i / 3.) * sin(angle - 0.05 + 0.1 * j / 4.)
                            self.game.entities.append(particle)

            if hasattr(self, 'last_fired_frame'):
                truc = frame - self.last_fired_frame
                if truc < 20 and truc % 4 == 0:
                    xdelta = 8 if self.game.player.direction else (24 - 8)
                    pos = (self.game.player.x + xdelta, self.game.player.y + 18)
                    bullet = UserBullet(pos)
                    bullet.speed[0] = self.game.player.speed[0]
                    bullet.speed[1] = -4.
                    self.game.entities.append(bullet)

            if frame > 70 and self.boss.life <= 0:
                self.change_level('test3')
                self.textbox.text = "You've won!"
            if self.game.player.life <= 0:
                self.change_level('test2')
                self.textbox.text = "You've just lost\n THE GAME!"
                self.game.player.image = loader.load_image('mad-scientist-dark.png')
                self.hamster = None
                self.frame = 410
        else:
            if frame == 100:
                self.textbox.text = ""



    def draw(self, frame):
        rect = self.get_screen_rect(self.zoomlevel)
        if self.parallax_background:
            ampli_bg = self.background.get_width() - WIDTH * self.zoomlevel
            ampli_parallax = self.parallax_background.get_width() - WIDTH * self.zoomlevel
            rect2 = rect[0] * ampli_parallax / ampli_bg, rect[1], rect[2], rect[3]
            screen.blit(self.parallax_background, (0, 0), rect2)
        screen.blit(self.background, (0, 0), rect)
        for entity in self.game.entities:
            entity.draw(screen, self.zoomlevel, rect)

        self.draw_hooks(frame)
        self.textbox.draw(screen, self.zoomlevel)
        if self.healthbar:
            self.healthbar.update(self.game.player.life)
            self.healthbar.draw(screen, self.zoomlevel)
        pygame.display.flip()


    def update(self, frame):
        self.update_hooks(frame)
        for entity in self.game.entities:
            try:
                mass = entity.mass
            except AttributeError:
                mass = 1.
            entity.speed[1] += .1 * mass
            entity.tick(frame)
        self.game.update(1.)


    def run(self):
        running = True
        clock = pygame.time.Clock()
        self.frame = 0
        while running:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    running = False
                if event.type == pygame.KEYDOWN:
                    if event.key in (pygame.K_q, pygame.K_ESCAPE):
                        running = False
                self.game.player.handle_event(event)
                self.event_hooks(event, self.frame)

            self.update(self.frame)
            self.draw(self.frame)

            clock.tick(50)

            self.frame += 1



if __name__ == '__main__':
    loader = Loader()
    zoomlevel = 4

    pygame.display.init()
    pygame.font.init()
    screen = pygame.display.set_mode((WIDTH * zoomlevel, HEIGHT * zoomlevel))
    pygame.display.set_caption("Please! This is science!")

    surface = pygame.Surface((WIDTH * zoomlevel, HEIGHT * zoomlevel), flags=pygame.SRCALPHA)

    game_runner = GameRunner(loader, zoomlevel)
    game_runner.run()

